import argparse
import os, random
from collections import Counter

from keras.models import Sequential
from keras.layers import Embedding, Bidirectional, LSTM, Dense, TimeDistributed

import numpy as np
from keras import backend as K
from tensorflow.keras import Model

from transformers import BertTokenizer, TFBertLMHeadModel, BertConfig, TFBertModel
import tensorflow as tf


'''
We'll use the same data as the Part-of-Speech tagger

Relevant data: 
Train: lm-data/little_prince_train.txt
Dev: lm-data/little_prince_dev.txt
Test: lm-data/little_prince_test.txt
'''

# default args

args = argparse.ArgumentParser(description='Program description.')
args.add_argument('-d','--device', default='cpu', help='Either "cpu" or "cuda"')
args.add_argument('-e','--epochs', default=20, type=int, help='Number of epochs')
args.add_argument('-lr','--learning-rate', default=0.1, type=float, help='Learning rate')
args.add_argument('-do','--dropout', default=0.3, type=int, help='Dropout rate')
args.add_argument('-em','--embedding-size', default=100, type=int, help='Embedding dimension size')
args.add_argument('-hs','--hidden-size', default=10, type=int, help='Hidden layer size')
args.add_argument('--use-bert', action='store_true', help='Use BERT tokenization and embeddings')
args.add_argument('-T','--train', type=str, help='Train file', required=True)
args.add_argument('-t','--test', type=str, help='Test or Dev file', required=True)
args.add_argument('-b','--batch-size', default=10, type=int, help='Batch Size')
args.set_defaults(use_bert=False)
args = args.parse_args()

train_file = args.train
dev_file = args.test

UNK = '[UNK]'
PAD = '[PAD]'
START = '<s>'
END = '</s>'

tokenizer = BertTokenizer.from_pretrained("bert-base-uncased")

def get_vocabulary_and_data_with_bert_tokenization(data_file):
    data = []
    with open(data_file, 'r', encoding='utf8') as f:
        for line in f:
            line = line.strip()
            if not line: continue
            sent = [START]
            sent.extend(tokenizer.tokenize(line))
            sent.append(END)
            data.append(sent)
    vocab = {k:v for k,v in tokenizer.vocab.items()}
    vocab['<s>'] = 101 # alias for [CLS]
    vocab['</s>'] = 102 # alias for [SEP]
    return vocab, data


def get_vocabulary_and_data(data_file, max_vocab_size=None, use_bert=False):
    if use_bert:
        return get_vocabulary_and_data_with_bert_tokenization(data_file)
    vocab = Counter()
    data = []
    with open(data_file, 'r', encoding='utf8') as f:
        for line in f:
            line = line.strip()
            if not line: continue
            sent = [START]
            tokens = transform_text_sequence(line.split())
            for tok in tokens:
                sent.append(tok)
                vocab[tok]+=1
            sent.append(END)
            data.append(sent)
            vocab[START]+=1
            vocab[END]+=1
    vocab = [w for w in sorted(vocab, key=lambda x:vocab[x], reverse=True)]
    if max_vocab_size:
        vocab = vocab[:max_vocab_size-2]
    vocab = [UNK, PAD] + vocab

    return {k:v for v,k in enumerate(vocab)}, data


def vectorize_sequence(seq, vocab):
    seq = [tok if tok in vocab else UNK for tok in seq]
    return [vocab[tok] for tok in seq]


def unvectorize_sequence(seq, vocab):
    translate = sorted(vocab.keys(),key=lambda k:vocab[k])
    return [translate[i] for i in seq]


def one_hot_encode_label(label, vocab):
    vec = [1.0 if l==label else 0.0 for l in vocab]
    return vec


def batch_generator_lm(data, vocab, batch_size=1):
    while True:
        batch_x = []
        batch_y = []
        for sent in data:
            batch_x.append(vectorize_sequence(sent, vocab))
            batch_y.append([one_hot_encode_label(token, vocab) for token in shift_by_one(sent)])
            if len(batch_x) >= batch_size:
                # Pad Sequences in batch to same length
                batch_x = pad_sequences(batch_x, vocab[PAD])
                batch_y = pad_sequences(batch_y, one_hot_encode_label(PAD, vocab))
                batch_x, batch_y = np.array(batch_x), np.array(batch_y)
                yield batch_x, batch_y.astype('float32')
                batch_x = []
                batch_y = []


def describe_data(data, generator):
    batch_x, batch_y = [], []
    for bx, by in generator:
        batch_x = bx
        batch_y = by
        break
    print('Data example:',data[49])
    print('Data size',len(data))
    print('Batch input shape:', batch_x.shape)
    print('Batch output shape:', batch_y.shape)


def pad_sequences(batch_x, pad_value):
    ''' This function should take a batch of sequences of different lengths
        and pad them with the pad_value token so that they are all the same length.

        Assume that batch_x is a list of lists.
    '''
    pad_length = len(max(batch_x, key=lambda x: len(x)))
    for i, x in enumerate(batch_x):
        if len(x) < pad_length:
            batch_x[i] = x + ([pad_value] * (pad_length - len(x)))

    return batch_x


def generate_text(language_model, vocab):
    prediction = [START]
    while not (prediction[-1] == END or len(prediction)>=50):
        next_token_one_hot = language_model.predict(np.array([[vocab[p] for p in prediction]]), batch_size=1)[0][-1]
        threshold = random.random()
        sum = 0
        next_token = 0
        for i,p in enumerate(next_token_one_hot):
            sum += p
            if sum>threshold:
                next_token = i
                break
        for w, i in vocab.items():
            if i==next_token:
                prediction.append(w)
                break
    return prediction


def perplexity(y_true, y_pred):
    # https://stackoverflow.com/questions/41881308/how-to-calculate-perplexity-of-rnn-in-tensorflow
    cross_entropy = K.categorical_crossentropy(y_true, y_pred)
    perp = K.exp(cross_entropy)
    return perp


class BERT_Wrapper(Model):

  def __init__(self):
    super(BERT_Wrapper, self).__init__()
    self.encoder = TFBertModel.from_pretrained("bert-base-uncased", trainable=False)
    self.dense = Dense(args.hidden_size)

  def call(self, inputs, **kwargs):
      outputs = self.encoder(inputs)
      last_hidden_states = outputs[0] # The last hidden-state is the first element of the output tuple
      output = self.dense(last_hidden_states)
      return output


# TODO
def load_pretrained_embeddings(glove_file, vocab):
    embedding_matrix = np.zeros((len(vocab), args.embedding_size))
    with open(glove_file, encoding='utf8') as f:
        for line in f:
            # Each line will be a word and a list of floats, separated by spaces.
            # If the word is in your vocabulary, create a numpy array from the list of floats.
            # Assign the array to the correct row of embedding_matrix.
            ...

    embedding_matrix[vocab[UNK]] = np.random.randn(args.embedding_size)
    return embedding_matrix

# TODO
def transform_text_sequence(seq):
    '''
    Implement this function if you want to transform the input text,
    for example normalizing case.
    '''
    return seq

# TODO
def shift_by_one(seq):
    '''
    input: ['<s>', 'The', 'dog', 'chased', 'the', 'cat', 'around', 'the', 'house', '</s>']
    output: ['The', 'dog', 'chased', 'the', 'cat', 'around', 'the', 'house', '</s>', '[PAD]']
    '''
    ...

def main():
    vocab, train_data = get_vocabulary_and_data(train_file, use_bert=args.use_bert)
    _, dev_data = get_vocabulary_and_data(dev_file, use_bert=args.use_bert)

    describe_data(train_data, batch_generator_lm(train_data, vocab, args.batch_size))


    # Implement your model here! ----------------------------------------------------------------------
    # Use the variables args.batch_size, args.hidden_size, args.embedding_size, args.dropout, args.epochs
    # You can input these as command line parameters.
    if args.use_bert:
        embedding_layer = ...
    else:
        embedding_layer = ...
    language_model = ...






    # ------------------------------------------------------------------------------------------------
    optimizer = tf.optimizers.Adadelta(clipvalue=0.5)
    language_model.compile(optimizer=optimizer, loss='categorical_crossentropy', metrics=['accuracy',perplexity])

    for i in range(args.epochs):
        print('Epoch',i+1,'/',args.epochs)
        # Training
        language_model.fit(batch_generator_lm(train_data, vocab, args.batch_size),
                                     epochs=1, steps_per_epoch=len(train_data)/args.batch_size)
        # Evaluation
        loss, acc, perp = language_model.evaluate(batch_generator_lm(dev_data, vocab),
                                                  steps=len(dev_data))
        print('Dev Loss:', loss, 'Dev Acc:', acc, 'Dev Peprlexity:', perp)

    for i in range(10):
        text = generate_text(language_model, vocab)
        print(text)


if __name__ == '__main__':
    main()